# Copyright 2012 - John Calixto
#
# This file is part of bang.
#
# bang is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# bang is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with bang.  If not, see <http://www.gnu.org/licenses/>.
"""
Customizations and enhancements of the upstream python-reddwarfclient library
to support hpcloud extensions.
"""
import reddwarfclient
import reddwarfclient.base


class DBSecurityGroup(reddwarfclient.base.Resource):
    def grant(self, source_cidr):
        """
        Allows connections from :attr:`source_cidr`.

        :param str source_cidr:  The source IP range from which to allow
            connections.

        """
        self.manager.api.secgroup_rules.create(self.id, source_cidr)


class DBSecurityGroups(reddwarfclient.base.ManagerWithFind):
    """Manages db security groups"""

    resource_class = DBSecurityGroup

    def list(self):
        """
        List all security groups.

        :rtype: :class:`list` of :class:`DBSecurityGroup`.

        """
        # that's right, the root key uses an underscore, not a hyphen.  the
        # hpcloud api docs show it as a hyphen.
        return self._list("/security-groups", "security_groups")


class DBSecurityGroupRule(reddwarfclient.base.Resource):
    def delete(self):
        """Deletes this rule"""
        self.manager.delete(self.id)


class DBSecurityGroupRules(reddwarfclient.base.ManagerWithFind):
    """Manages db security group rules"""

    resource_class = DBSecurityGroupRule

    def create(self, dbsecgroup_id, source_cidr, port=3306):
        """
        Creates a security group rule.

        :param str dbsecgroup_id:  The ID of the security group in which this
            rule should be created.
        :param str source_cidr:  The source IP address range from which access
            should be allowed.
        :param int port:  The port number used by db clients to connect to the
            db server.  This would have been specified at db instance creation
            time.

        :rtype:  :class:`DBSecurityGroupRule`.

        """
        body = {
                "security_group_rule": {
                    "security_group_id": dbsecgroup_id,
                    "cidr": source_cidr,
                    "from_port": port,
                    "to_port": port,
                    }
                }
        return self._create("/security-group-rules", body,
                "security_group_rule")

    def delete(self, rule_id):
        """
        Deletes a rule from a security group.

        :param str rule_id:  The rule ID.

        """
        self._delete("/security-group-rules/%s" % rule_id)


class HPDbaas(reddwarfclient.Dbaas):
    def __init__(self, *args, **kwargs):
        super(HPDbaas, self).__init__(*args, **kwargs)
        self.secgroups = DBSecurityGroups(self)
        self.secgroup_rules = DBSecurityGroupRules(self)



# Copyright 2012 - John Calixto
#
# This file is part of bang.
#
# bang is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# bang is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with bang.  If not, see <http://www.gnu.org/licenses/>.
import os.path

try:
    import bang.version
    VERSION = bang.version.VERSION
except ImportError:
    VERSION = 'badonkadonk'

BANG_DIR = os.path.abspath(
        os.path.join(
            os.path.dirname(__file__),
            '..',
            )
        )

WORK_DIR = os.path.join(BANG_DIR, 'work')

CONFIG_DIR = os.path.join(BANG_DIR, 'config')


class BangError(Exception):
    pass


class TimeoutError(BangError):
    pass



# Copyright 2012 - John Calixto
#
# This file is part of bang.
#
# bang is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# bang is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with bang.  If not, see <http://www.gnu.org/licenses/>.
import random
import string

# at least RDS appears to force lowercase even if you pass in mixed case
_AWS_NAME_CHARS = string.lowercase + string.digits


class Provider(object):
    """The base class for all providers."""
    def __init__(self, creds):
        self.creds = creds

        # Minimal attempt to prevent obvious postfix duplication
        self.component_names = []

    def gen_component_name(self, basename, postfix_length=13):
        """
        Creates a resource identifier with a random postfix.  This is an
        attempt to minimize name collisions in provider namespaces.

        :param str basename:  The string that will be prefixed with the stack
            name, and postfixed with some random string.

        :param int postfix_length:  The length of the postfix to be appended.

        """
        def newcname():
            postfix = ''.join(
                    random.choice(_AWS_NAME_CHARS)
                    for i in xrange(postfix_length)
                    )
            return '%s-%s' % (basename, postfix)
        cname = newcname()
        while cname in self.component_names:
            cname = newcname()
        self.component_names.append(cname)
        return cname

    def get_consul(self, resource_type):
        """
        Returns an object that a :class:`~bang.deployers.deployer.Deployer`
        uses to control resources of :attr:`resource_type`.

        :param str service:  Any of the resources defined
            in :mod:`bang.resources`.

        """
        consul = self.CONSUL_MAP.get(resource_type)
        if consul:
            return consul(self)


class Consul(object):
    """
    The base class for all service consuls.

    Not really the boss of anything, but conveys intent-from-above to foreign
    entities (e.g. OpenStack Nova/Swift, AWS EC2/S3/RDS, etc...).  Also
    communicates the state of the world back up to the boss.

    """

    def __init__(self, provider):
        self.provider = provider



# Copyright 2012 - John Calixto
#
# This file is part of bang.
#
# bang is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# bang is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with bang.  If not, see <http://www.gnu.org/licenses/>.
import copy
import json
import multiprocessing
import os.path

from ansible.callbacks import (
        AggregateStats,
        PlaybookCallbacks,
        PlaybookRunnerCallbacks,
        )
from ansible.playbook import PlayBook
from .deployers import get_stage_deployers
from .inventory import BangsibleInventory
from .util import log, SharedNamespace, SharedMap
from . import BangError, resources as R, attributes as A


class Stack(object):
    """
    Deploys infrastructure/platform resources, then configures any deployed
    servers using ansible playbooks.

    """
    def __init__(self, config):
        """
        :param config:  A mapping object with configuration keys and values.
            May be arbitrarily nested.
        :type config:  :class:`bang.config.Config`

        """
        self.name = config[A.NAME]
        self.version = config[A.VERSION]
        self.config = config
        self.manager = multiprocessing.Manager()
        self.shared_namespaces = {}

        self.groups_and_vars = SharedMap(self.manager)
        """
        Deployers stash inventory data for any newly-created servers in this
        mapping object.  Note: uses SharedMap because this must be
        multiprocess-safe.

        """

        # TODO: suss out autoscaling. see count_to_deploy()

    def get_deployers(self):
        """
        Returns a :class:`list` of *stages*, where each *stage* is a
        :class:`list` of :class:`~bang.deployers.deployer.Deployer` objects.
        It defines the execution order of the various deployers.

        """
        return enumerate(
                [get_stage_deployers(keys, self) for keys in R.STAGES]
                )

    def get_namespace(self, key):
        """
        Returns a :class:`~bang.util.SharedNamespace` for the given
        :attr:`key`.  These are used by
        :class:`~bang.deployers.deployer.Deployer` objects of the same
        ``deployer_class`` to coordinate control over multiple deployed
        instances of like resources.  E.g. With 5 clones of an application
        server, 5 :class:`~bang.deployers.deployer.Deployer` objects in
        separate, concurrent processes will use the same shared namespace to
        ensure that each object/process controls a distinct server.

        :param str key:  Unique ID for the
            namespace.  :class:`~bang.deployers.deployer.Deployer` objects that
            call :meth:`get_namespace` with the same :attr:`key` will receive
            the same :class:`~bang.util.SharedNamespace` object.

        """
        namespace = self.shared_namespaces.get(key)
        if namespace:
            return namespace
        ns = SharedNamespace(self.manager)
        self.shared_namespaces[key] = ns
        return ns

    def find_first(self, attr_name, resources, extra_prefix=''):
        """
        Returns the boto object for the first resource in ``resources`` that
        belongs to this stack.  Uses the attribute specified by ``attr_name``
        to match the stack name.

        E.g.  An RDS instance for a stack named ``foo`` might be named
        ``foo-mydb-fis8932ifs``.  This call::

            find_first('id', conn.get_all_dbinstances())

        would return the boto.rds.dbinstance.DBInstance object whose ``id`` is
        ``foo-mydb-fis8932ifs``.

        Returns None if a matching resource is not found.

        If specified, ``extra_prefix`` is appended to the stack name prefix
        before matching.
        """
        prefix = self.name + '-' + (extra_prefix + '-' if extra_prefix else '')
        for res in resources:
            attr = getattr(res, attr_name)
            if attr.startswith(prefix):
                return res

    def add_host(self, host, group_names=None, host_vars=None):
        """
        Used by deployers to add hosts to the inventory.

        :param str host:  The host identifier (e.g. hostname, IP address) to
            use in the inventory.

        :param list group_names:  A list of group names to which the host
            belongs.  **Note:  This list will be sorted in-place.**

        :param dict host_vars:  A mapping object of host *variables*.  This can
            be a nested structure, and is used as the source of all the
            variables provided to the ansible playbooks.  **Note:  Additional
            key-value pairs (e.g. dynamic ansible values like
            ``inventory_hostname``) will be inserted into this mapping
            object.**

        """
        gnames = group_names if group_names else []
        hvars = host_vars if host_vars else {}

        # Add in ansible's magic variables.  Assign them here because this is
        # just about the earliest point we can calculate them before anything
        # ansible-related (e.g. Stack.configure(), ``bang --host``) executes.
        gnames.sort()
        hvars[A.server.GROUP_NAMES] = gnames
        hvars[A.server.INV_NAME] = host
        hvars[A.server.INV_NAME_SHORT] = host.split('.')[0]

        self.groups_and_vars.merge(host, hvars)

        for gname in group_names:
            self.groups_and_vars.append(gname, host)

    def describe(self):
        """Iterates through the deployers but doesn't run anything"""
        for stage, corunners in self.get_deployers():
            print self.name, "STAGE ", stage
            for d in corunners:
                print d.__class__.__name__, ",".join(
                        [p[1].__name__ for p in d.phases]
                        )

    def _run(self, action):
        for stage, corunners in self.get_deployers():
            children = []
            errors = 0
            for d in corunners:
                p = multiprocessing.Process(
                        name=d.__class__.__name__,
                        target=d.run,
                        args=(action, ),
                        )
                children.append(p)
                p.start()
            for child in children:
                child.join()
                if child.exitcode != 0:
                    errors += 1
            if errors:
                msg = "Stage %d had %d errors." % (stage, errors)
                log.error(msg)
                raise BangError(msg)

    def deploy(self):
        """
        Iterates through the deployers returned by ``self.get_deployers()``.

        Deployers in the same stage are run concurrently.  The runner only
        proceeds to the next stage once all of the deployers in the same
        stage have completed successfully.

        Any failures in a stage cause the run to terminate before proceeding to
        the next stage.

        """
        self._run('deploy')

    def configure(self):
        """
        Executes the ansible playbooks that configure the servers in the stack.

        Assumes that the root playbook directory is ``./playbooks/`` relative
        to the stack configuration file.  Also sets the ansible *module_path*
        to be ``./common_modules/`` relative to the stack configuration file.

        E.g.  If the stack configuration file is::

            $HOME/bang-stacks/my_web_service.yml

        then the root playbook directory is::

            $HOME/bang-stacks/playbooks/

        and the ansible module path is::

            $HOME/bang-stacks/common_modules/

        """
        bang_config_dir = os.path.abspath(
                os.path.dirname(self.config.filepath)
                )
        ansible_module_dir = os.path.join(bang_config_dir, 'common_modules')
        playbook_dir = os.path.join(bang_config_dir, 'playbooks')
        for playbook in self.config.get(A.PLAYBOOKS, []):
            playbook_path = os.path.join(playbook_dir, playbook)

            # gratuitously stolen from main() in ``ansible-playbook``
            stats = AggregateStats()
            playbook_cb = PlaybookCallbacks(verbose=1)
            runner_cb = PlaybookRunnerCallbacks(stats, verbose=1)

            pb = PlayBook(
                playbook=playbook_path,
                module_path=ansible_module_dir,

                # ``host_list`` is used to generate the inventory, but don't
                # worry, we override the inventory later
                host_list=[],

                # TODO: determine forks
                # forks=options.forks,

                callbacks=playbook_cb,
                runner_callbacks=runner_cb,
                stats=stats,
            )
            pb.inventory = BangsibleInventory(
                    copy.deepcopy(self.groups_and_vars.lists),
                    copy.deepcopy(self.groups_and_vars.dicts),
                    )

            pb.run()

            hosts = sorted(pb.stats.processed.keys())
            failed = False
            for h in hosts:
                hsum = pb.stats.summarize(h)
                if hsum['failures'] or hsum['unreachable']:
                    failed = True
                print "%-30s : %s" % (h, hsum)
                # TODO: sort this out
                # print "%-30s : %s %s %s %s " % (
                #     hostcolor(h, hsum),
                #     colorize('ok', hsum['ok'], 'green'),
                #     colorize('changed', hsum['changed'], 'yellow'),
                #     colorize('unreachable', hsum['unreachable'], 'red'),
                #     colorize('failed', hsum['failures'], 'red'))

            if failed:
                raise BangError("Server configuration failed!")

    def show_inventory(self):
        """
        Satisfies the ``--list`` portion of ansible's external inventory API.

        Allows ``bang`` to be used as an external inventory script, for example
        when running ad-hoc ops tasks.  For more details, see:
        http://ansible.cc/docs/api.html#external-inventory-scripts

        """
        self._run('inventory')
        print json.dumps(copy.deepcopy(self.groups_and_vars.lists))

    def show_host(self, host):
        """
        Satisfies the ``--host`` portion of ansible's external inventory API.

        Allows ``bang`` to be used as an external inventory script, for example
        when running ad-hoc ops tasks.  For more details, see:
        http://ansible.cc/docs/api.html#external-inventory-scripts

        """
        self._run('inventory')
        # use a deepcopy to make sure data doesn't move around on us in
        # multiprocess-land:
        all_hvars = copy.deepcopy(self.groups_and_vars.dicts)
        hostvars = all_hvars.get(host, {})
        print json.dumps(hostvars)



# Copyright 2012 - John Calixto
#
# This file is part of bang.
#
# bang is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# bang is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with bang.  If not, see <http://www.gnu.org/licenses/>.
import ansible.inventory
from ansible.inventory.group import Group
from ansible.inventory.host import Host


def get_ansible_groups(group_map):
    """
    Constructs a list of :class:`ansible.inventory.group.Group` objects from a
    map of lists of host strings.

    """
    # Much of this logic is cribbed from
    # ansible.inventory.script.InventoryScript
    all_hosts = {}
    group_all = Group('all')
    groups = [group_all]
    for gname, hosts in group_map.iteritems():
        g = Group(gname)
        for host in hosts:
            h = all_hosts.get(host, Host(host))
            all_hosts[host] = h
            g.add_host(h)
            group_all.add_host(h)
        group_all.add_child_group(g)
        groups.append(g)
    return groups


class BangsibleInventory(ansible.inventory.Inventory):
    def __init__(self, groups, hostvars):

        # caching to avoid repeated calculations, particularly with external
        # inventory scripts.
        self._vars_per_group = {}
        self._hosts_cache = {}
        self._groups_list = {}

        # a list of host(names) to contain current inquiries to
        self._restriction = None
        self._also_restriction = None
        self._subset = None

        # whether the inventory file is a script
        self._is_script = False

        self.groups = get_ansible_groups(groups)

        # Prepopulate the cache.  The base Inventory only gathers host vars as
        # necessary (to avoid repeated execs of the inventory script), and
        # caches them in ``self._vars_per_host``.  By the time bang needs a
        # BangsibleInventory, it already has all of the hostvars so we just set
        # the cache to be the hostvars dict.
        self._vars_per_host = hostvars

    def is_file(self):
        return False



# Copyright 2012 - John Calixto
#
# This file is part of bang.
#
# bang is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# bang is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with bang.  If not, see <http://www.gnu.org/licenses/>.
import collections
import os
import os.path
import tempfile
import yaml

from . import resources as R, attributes as A
from .util import log, bump_version_tail, deep_merge_dicts


DEFAULT_CONFIG_DIR = 'bang-stacks'
DEFAULT_LAUNCH_TIMEOUT_S = 0

RC_KEYS = [A.DEPLOYER_CREDS, 'config_dir']
ALL_RESERVED_KEYS = RC_KEYS + R.DYNAMIC_RESOURCE_KEYS


def find_component_tarball(bucket, comp_name, comp_config):
    """
    Returns True if the component tarball is found in the bucket.

    Otherwise, returns False.
    """
    values = {
            'name': comp_name,
            'version': comp_config['version'],
            'platform': comp_config['platform'],
            }
    template = comp_config.get('archive_template')
    if template:
        key_name = template % values
    else:
        key_name = '%(name)s/%(name)s-%(version)s.tar.gz' % values
    if not bucket.get_key(key_name):
        log.error('%s not found' % key_name)
        return False
    return True


def parse_bangrc():
    """
    Parses ``$HOME/.bangrc`` for global settings and deployer credentials.  The
    ``.bangrc`` file is expected to be a YAML file whose outermost structure is
    a key-value map.

    Note that even though ``.bangrc`` is just a YAML file in which a user could
    store any top-level keys, it is not expected to be used as a holder of
    default values for stack-specific configuration attributes - if present,
    they will be ignored.

    Returns {} if ``$HOME/.bangrc`` does not exist.

    :rtype:  :class:`dict`

    """
    bangrc_path = os.path.join(os.environ['HOME'], '.bangrc')
    try:
        with open(bangrc_path) as f:
            raw = yaml.load(f)
        return dict((k, raw[k]) for k in raw if k in RC_KEYS)
    except IOError:
        pass
    return {}


def resolve_config_spec(config_spec, config_dir=''):
    """
    Resolves :attr:`config_spec` to a path to a config file.

    :param str config_spec:  Valid config specs:

        - The basename of a YAML config file *without* the ``.yml`` extension.
          The full path to the config file is resolved by appending ``.yml`` to
          the basename, then by searching for the result in the
          :attr:`config_dir`.

        - The path to a YAML config file.  The path may be absolute or may be
          relative to the current working directory.  If :attr:`config_spec`
          contains a ``/`` (forward slash), or if it ends in ``.yml``, it is
          treated as a path.

    :param str config_dir:  The directory in which to search for stack
        configuration files.

    :rtype:  :class:`str`

    """
    if '/' in config_spec or config_spec.endswith('.yml'):
        return config_spec
    return os.path.join(config_dir, '%s.yml' % config_spec)


class Config(dict):
    """
    A dict-alike that provides a convenient constructor, stashes the path
    to the config file as an instance attribute, and performs some validation
    of the values.
    """
    def __init__(self, *args, **kwargs):
        """

        :param str path_to_yaml:  Path to a yaml file to use as the data source
            for the returned instance.

        """
        super(Config, self).__init__(*args, **kwargs)
        self.filepath = ''

    @classmethod
    def from_config_specs(cls, config_specs, prepare=True):
        """
        Alternate constructor that merges config attributes from
        ``$HOME/.bangrc`` and :attr:`config_specs` into a single
        :class:`Config` object.

        The first (and potentially *only* spec) in :attr:`config_specs` should
        be main configuration file for the stack to be deployed.  The returned
        object's :attr:`filepath` will be set to the absolute path of the first
        config file.

        If multiple config specs are supplied, their values are merged together
        in the order specified in :attr:`config_specs` - That is, later values
        override earlier values.

        :param config_specs:  List of config specs.
        :type config_specs:  :class:`list` of :class:`str`

        :param bool prepare:  Flag to control whether or not :meth:`prepare` is
            called automatically before returning the object.

        :rtype:  :class:`Config`

        """
        bangrc = parse_bangrc()
        config_dir = bangrc.get('config_dir', DEFAULT_CONFIG_DIR)
        config_paths = [
                resolve_config_spec(cs, config_dir) for cs in config_specs
                ]
        config = cls()
        config.update(bangrc)
        if config_paths:
            config.filepath = config_paths[0]
        for c in config_paths:
            with open(c) as f:
                deep_merge_dicts(config, yaml.load(f))
        if prepare:
            config.prepare()
        return config

    def _prepare_dbs(self):
        dbcreds = self.get(R.DATABASE_CREDS, {})
        for db in self.get(R.DATABASES, []):
            name = db[A.database.NAME]

            # default db name is the deployer instance name
            if A.database.DB_NAME not in db:
                db[A.database.DB_NAME] = name
            if A.database.LAUNCH_TIMEOUT not in db:
                db[A.database.LAUNCH_TIMEOUT] = DEFAULT_LAUNCH_TIMEOUT_S

            # make global db creds available to db deployer instance
            creds = dbcreds.get(name)
            if not creds:
                continue
            db[A.database.ADMIN_USER] = creds[A.database.ADMIN_USER]
            db[A.database.ADMIN_PASS] = creds[A.database.ADMIN_PASS]

    def _prepare_secgroups(self):
        stack = self[A.NAME]

        sg_rule_sets = []
        for sg in self.get(R.SERVER_SECURITY_GROUPS, []):
            provider = sg[A.secgroup.PROVIDER]
            region = sg[A.secgroup.REGION]
            # dress up the stack secgroup names
            dressy_name = '%s-%s' % (stack, sg[A.secgroup.NAME])
            sg[A.secgroup.NAME] = dressy_name
            rules = []
            for rule in sg[A.secgroup.RULES]:
                r = rule.copy()

                # generate dressy source if necessary
                if rule.get(A.secgroup.SOURCE_SELF):
                    r[A.secgroup.SOURCE] = dressy_name

                r[A.secgroup.TARGET] = dressy_name
                rules.append(r)
            sg_rule_sets.append(
                    {
                        A.secgroup.NAME: dressy_name,
                        A.secgroup.PROVIDER: provider,
                        A.secgroup.REGION: region,
                        A.secgroup.RULES: rules,
                        }
                    )
        self[R.SERVER_SECURITY_GROUP_RULES] = sg_rule_sets

        # make sure server configs use the dressed-up secgroup names
        for s in self.get(R.SERVERS, []):
            groups = [
                    '%s-%s' % (stack, gname)
                    for gname in s.get(A.server.STACK_SECGROUPS, [])
                    ]
            groups.extend(s.get(A.server.EXTRA_SECGROUPS, []))
            s[A.server.SECGROUPS] = groups

    def _prepare_tags(self):
        # add ``stack`` and ``role`` tags
        for s in self.get(R.SERVERS, []):
            tags = s.get(A.server.TAGS, {})
            tags[A.tags.STACK] = self[A.NAME]
            tags[A.tags.ROLE] = s[A.server.NAME]
            s[A.server.TAGS] = tags

    def _prepare_ssh_keys(self):
        keys = self.get(A.DEPLOYER_CREDS, {}).get(R.SSH_KEYS, {})
        if not keys:
            return
        keys_to_install = []
        for server in self.get(R.SERVERS, []):
            key_name = server.get(A.server.SSH_KEY)
            if not key_name:
                continue
            key = keys.get(key_name)
            if not key:
                continue
            keys_to_install.append(
                    {
                        A.ssh_key.NAME: key_name,
                        A.ssh_key.KEY: key,
                        A.ssh_key.PROVIDER: server[A.server.PROVIDER],
                        A.ssh_key.REGION: server[A.server.REGION],
                        }
                    )
        if keys_to_install:
            self[R.SSH_KEYS] = keys_to_install

    def _prepare_servers(self):
        """
        Prepare the variables that are exposed to the servers.

        Most attributes in the server config are used directly.  However, due
        to variations in how cloud providers treat regions and availability
        zones, this method allows either the ``availability_zone`` or the
        ``region_name`` to be used as the target availability zone for a
        server.  If both are specified, then ``availability_zone`` is used.  If
        ``availability_zone`` is not specified in the server config, then the
        ``region_name`` value is used as the target availability zone.

        """
        stack = {
                A.NAME: self[A.NAME],
                A.VERSION: self[A.VERSION],
                }
        for server in self.get(R.SERVERS, []):
            # default values
            if A.server.LAUNCH_TIMEOUT not in server:
                server[A.server.LAUNCH_TIMEOUT] = DEFAULT_LAUNCH_TIMEOUT_S
            if A.server.POST_DELAY not in server:
                server[A.server.POST_DELAY] = DEFAULT_LAUNCH_TIMEOUT_S
            if A.server.AZ not in server:
                server[A.server.AZ] = server[A.server.REGION]

            # distribute the config scope attributes
            svars = {A.STACK: stack}
            for scope in server.get(A.server.SCOPES, []):
                # allow scopes to be defined inline
                if isinstance(scope, collections.Mapping):
                    svars.update(scope)
                else:
                    svars[scope] = self[scope]
            server[A.server.VARS] = svars

    def prepare(self):
        """
        Reorganizes the data such that the deployment logic can find it all
        where it expects to be.

        The raw configuration file is intended to be as human-friendly as
        possible partly through the following mechanisms:

            - In order to minimize repetition, any attributes that are common
              to all server configurations can be specified in the
              ``server_common_attributes`` stanza even though the stanza itself
              does not map directly to a deployable resource.
            - For reference locality, each security group stanza contains its
              list of rules even though rules are actually created in a
              separate stage from the groups themselves.

        In order to make the :class:`Config` object more useful to the program
        logic, this method performs the following transformations:

            - Distributes the ``server_common_attributes`` among all the
              members of the ``servers`` stanza.
            - Extracts security group rules to a top-level key, and
              interpolates all source and target values.

        """
        # TODO: take server_common_attributes and disperse it among the various
        # server stanzas
        self._prepare_ssh_keys()
        self._prepare_secgroups()
        self._prepare_tags()
        self._prepare_dbs()
        self._prepare_servers()

    def validate(self):
        """
        Performs all validation checks on this config.

        Raises :class:`ValueError` for invalid configs.

        """
        # TODO: validate providers in each of the resource stanzas

    def autoinc(self):
        """
        Conditionally updates the stack version in the file associated with
        this config.

        This  handles both official releases (i.e. QA configs), and release
        candidates.  Assumptions about version:

            - Official release versions are MAJOR.minor, where MAJOR and minor
              are both non-negative integers.  E.g.

                2.9
                2.10
                2.11
                3.0
                3.1
                3.2
                etc...

            - Release candidate versions are MAJOR.minor-rc.N, where MAJOR,
              minor, and N are all non-negative integers.

                3.5-rc.1
                3.5-rc.2

        """
        if not self.get('autoinc_version'):
            return

        oldver = self['version']
        newver = bump_version_tail(oldver)

        config_path = self.filepath
        temp_fd, temp_name = tempfile.mkstemp(
                dir=os.path.dirname(config_path),
                )
        with open(config_path) as old, os.fdopen(temp_fd, 'w') as new:
            for oldline in old:
                if oldline.startswith('version:'):
                    new.write("version: '%s'\n" % newver)
                    continue
                new.write(oldline)

        # no need to backup the old file, it's under version control anyway -
        # right???
        log.info('Incrementing stack version %s -> %s' % (oldver, newver))
        os.rename(temp_name, config_path)
        # TODO: commit_bang_config()



# Copyright 2012 - John Calixto
#
# This file is part of bang.
#
# bang is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# bang is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with bang.  If not, see <http://www.gnu.org/licenses/>.
import argparse
import atexit
import bang
import collections
import copy
import json
import logging
import multiprocessing
import time
import re
import subprocess
import sys
import textwrap
from datetime import datetime
from logging.handlers import BufferingHandler

import boto
from boto.s3.key import Key
from logutils.queue import QueueHandler, QueueListener


CONSOLE_LOGGING_FORMAT = '%(asctime)s %(levelname)8s %(processName)s - %(message)s'

# use the multiprocessing logger so when we start parallelizing the
# deploys, we have a seamless transition.
_mlog = multiprocessing.get_logger()
log = _mlog.getChild('bang')

# give this logger at least one handler to avoid pesky warnings for bang
# commands that don't actually care about logging
_mlog.addHandler(logging.NullHandler())
log.addHandler(logging.NullHandler())
del _mlog


class StrictAttrBag(object):
    """
    Generic attribute container that makes constructor arguments available as
    object attributes.

    Checks :meth:`__init__` argument names against lists of *required* and
    *optional* attributes.

    """
    def __init__(self, **kwargs):
        for reqk in self.REQUIRED_ATTRS:
            if reqk not in kwargs:
                raise TypeError("Missing required argument, %s" % reqk)
            setattr(self, reqk, kwargs.pop(reqk))

        for k, v in kwargs.items():
            if k not in self.OPTIONAL_ATTRS:
                raise TypeError("Unknown argument, %s" % k)
            setattr(self, k, v)


class SharedMap(object):
    """
    A multiprocess-safe :class:`Mapping` object that can be used to return
    values from child processes.

    """
    def __init__(self, manager):
        self.lists = manager.dict()
        self.dicts = manager.dict()
        self.lock = multiprocessing.Lock()

    def append(self, list_name, value):
        """Appends :attr:`value` to the list named :attr:`list_name`."""
        with self.lock:
            l = self.lists.get(list_name)
            if l:
                l.append(value)
            else:
                l = [value]
            self.lists[list_name] = l

    def merge(self, dict_name, values):
        """
        Performs deep-merge of :attr:`values` onto the :class:`Mapping` object
        named :attr:`dict_name`.

        If :attr:`dict_name` does not yet exist, then a deep copy of
        :attr:`values` is assigned as the initial mapping object for the given
        name.

        :param str dict_name:  The name of the dict onto which the values
            should be merged.

        """

        with self.lock:
            d = self.dicts.get(dict_name)
            if d:
                deep_merge_dicts(d, values)
            else:
                d = copy.deepcopy(values)
            self.dicts[dict_name] = d


class SharedNamespace(object):
    """
    A multiprocess-safe namespace that can be used to coordinate naming similar
    resources uniquely.  E.g. when searching for existing nodes in a cassandra
    cluster, you can use this SharedNamespace to make sure other processes
    aren't looking at the same node.
    """
    def __init__(self, manager):
        self.names = manager.list()
        self.lock = multiprocessing.Lock()

    def add_if_unique(self, name):
        """
        Returns ``True`` on success.

        Returns ``False`` if the name already exists in the namespace.
        """
        with self.lock:
            if name not in self.names:
                self.names.append(name)
                return True
        return False


class JSONFormatter(logging.Formatter):
    def __init__(self, config):
        logging.Formatter.__init__(self)
        self.stack = config.get('stack', 'anonymous')

    def format(self, record):
        timestamp = datetime.fromtimestamp(record.created).isoformat()
        out = {
                'level': record.levelname,
                'message': record.message,
                'timestamp': timestamp,
                'stack': self.stack,
                'pid': record.process,
                'process_name': record.processName,
                }
        return '%s\n' % json.dumps(out)


class ColoredConsoleFormatter(logging.Formatter):
    def format(self, record):
        pre = logging.Formatter.format(self, record)
        return '\033[38;5;%dm%s\033[0;37m' % (
                (record.process * 4) % 210 + 20,
                pre,
                )


class S3Handler(BufferingHandler):
    """
    Buffers all logging events, then uploads them all at once "atexit" to a
    single file in S3.
    """
    def __init__(self, bucket, prefix=''):
        BufferingHandler.__init__(self, 0)
        self.bucket = bucket
        self.prefix = prefix

    def shouldFlush(self, record):
        return False

    def flush(self):
        payload = ''
        while len(self.buffer) > 0:
            record = self.buffer.pop(0)
            if record.levelno >= self.level:
                payload += self.format(record)
        if payload:
            conn = boto.connect_s3()
            bucket = conn.get_bucket(self.bucket)
            key = Key(bucket)
            key.key = '/'.join((self.prefix, 'bang-%f' % time.time()))
            key.content_type = 'application/json'
            key.set_contents_from_string(payload)


def initialize_logging(config):
    multiprocessing.current_process().name = 'Stack'
    cfg = config.get('logging', {})
    console_level = cfg.get('console_level', 'INFO')
    log.setLevel(console_level)

    # log to s3 if there's a destination specified in the config
    bucket = cfg.get('s3_bucket')
    if bucket:
        json_formatter = JSONFormatter(config)
        s3_handler = S3Handler(bucket, cfg.get('s3_prefix', ''))
        s3_handler.setFormatter(json_formatter)
        s3_handler.setLevel(logging.INFO)

        # The parent process is the only one that actually buffers the log
        # records in memory and writes them out to s3.  The child processes
        # send all of their log records to the parent's queue.
        #
        # Using the QueueHandler and QueueListener classes from logutils-0.3.2
        # here since they're the implementations in future versions of stdlib
        # logging anyway (logutils is the "backports from Py3k logging"
        # library).
        queue = multiprocessing.Queue()
        ql = QueueListener(queue, s3_handler)

        def cleanup():
            ql.stop()
            s3_handler.flush()
        atexit.register(cleanup)
        ql.start()

        qh = QueueHandler(queue)
        log.addHandler(qh)

    # also log to stderr
    if sys.stderr.isatty():
        formatter = ColoredConsoleFormatter(CONSOLE_LOGGING_FORMAT)
    else:
        formatter = logging.Formatter(CONSOLE_LOGGING_FORMAT)
    handler = logging.StreamHandler()  # default stream is stderr
    handler.setFormatter(formatter)
    handler.setLevel(logging.DEBUG)
    log.addHandler(handler)
    log.debug('Logging initialized.')


def poll_with_timeout(timeout_s, break_func, wake_every_s=60):
    """
    Calls :attr:`break_func` every :attr:`wake_every_s` seconds for a total
    duration of :attr:`timeout_s` seconds, or until :attr:`break_func` returns
    something other than ``None``.

    If :attr:`break_func` returns anything other than ``None``, that value is
    returned immediately.

    Otherwise, continues polling until the timeout is reached, then returns
    ``None``.

    """
    time_slept = 0
    if wake_every_s > 60:
        msg = '... sleeping for %0.2f minutes' % (wake_every_s / 60.0)
    else:
        msg = '... sleeping for %d seconds' % wake_every_s
    res = break_func()
    while res is None and time_slept < timeout_s:
        log.debug(msg)
        time.sleep(wake_every_s)
        time_slept += wake_every_s
        res = break_func()
    return res


def parse_args(arg_config, alt_args=None):
    """
    ``alt_args`` is an optional list of strings to use *instead* of sys.argv.
    """
    parser = argparse.ArgumentParser(
            prog=arg_config.get('prog'),
            description=textwrap.dedent(arg_config['description']),
            formatter_class=argparse.RawTextHelpFormatter,
            )
    for ac in arg_config['arguments']:
        args = ac[:-1]
        kwargs = ac[-1]
        parser.add_argument(*args, **kwargs)
    return parser.parse_args(alt_args)


SECRET_PATTERN = re.compile(r'(\s*(\S*(password|pwd|key|secret|god)\S*)\s*:\s*)\S+')
SECRET_WHITELIST = ('ssh_key', 'key_pair')


def redact_secrets(line):
    """
    Returns a sanitized string for any ``line`` that looks like it contains a
    secret (i.e. matches SECRET_PATTERN).
    """
    def redact(match):
        if match.group(2) in SECRET_WHITELIST:
            return match.group(0)
        return match.group(1) + 'TOO_TOO_SEXY'
    return SECRET_PATTERN.sub(redact, line)


def bump_version_tail(oldver):
    """
    Takes any dot-separated version string and increments the rightmost field
    (which it expects to be an integer).
    """
    head, tail = oldver.rsplit('.', 1)
    return '%s.%d' % (head, (int(tail) + 1))


def count_to_deploy(stack, descriptor, config_count):
    """
    takes the max of config_count and number of instances running
    with this stack/descriptor combo
    """
    live_count = count_by_tag(stack, descriptor)
    if config_count > live_count:
        live_count = config_count

    return live_count


def count_by_tag(stack, descriptor):
    """
    Returns the count of currently running or pending instances
    that match the given stack and deployer combo
    """
    ec2_conn = boto.ec2.connection.EC2Connection()
    resses = ec2_conn.get_all_instances(
                    filters={
                        'tag:stack': stack,
                        'tag:descriptor': descriptor
                    })
    instance_list_raw = list()
    [[instance_list_raw.append(x) for x in res.instances] for res in resses]
    instance_list = [x for x in instance_list_raw if state_filter(x)]
    instances = len(instance_list)
    return instances


def state_filter(instance):
    """
    Helper function for count_by_tag
    """
    if instance.state == 'running' or instance.state == 'pending':
        return True
    else:
        return False


def deep_merge_dicts(base, incoming):
    """
    Performs an *in-place* deep-merge of key-values from :attr:`incoming`
    into :attr:`base`.  No attempt is made to preserve the original state of
    the objects passed in as arguments.

    :param dict base:  The target container for the merged values.  This will
        be modified *in-place*.
    :type base:  Any :class:`dict`-like object

    :param dict incoming:  The container from which incoming values will be
        copied.  Nested dicts in this will be modified.
    :type incoming:  Any :class:`dict`-like object

    :rtype:  None

    """
    for ki, vi in incoming.iteritems():
        if (ki in base
                and isinstance(vi, collections.MutableMapping)
                and isinstance(base[ki], collections.MutableMapping)
                ):
            deep_merge_dicts(base[ki], vi)
        else:
            base[ki] = vi


def fork_exec(cmd_list, input_data=None):
    """
    Like the subprocess.check_*() helper functions, but tailored to bang.

    ``cmd_list`` is the command to run, and its arguments as a list of strings.

    ``input_data`` is the optional data to pass to the command's stdin.

    On success, returns the output (i.e. stdout) of the remote command.

    On failure, raises BangError with the command's stderr.
    """
    # log.debug('fork_exec: cmd_list = %s, input_data = ^%s^' %
    #         (cmd_list, input_data))
    p = subprocess.Popen(
            cmd_list,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            )
    out = p.communicate(input_data)
    if p.returncode != 0:
        raise bang.BangError('ret: %d, stdout: ^%s^, stderr: ^%s^' %
                (p.returncode, out[0], out[1]))
    return out[0]



import urllib
from novaclient import base
from novaclient.v1_1.volumes import VolumeManager


VOLUMES_ENDPOINT = '/os-volumes'


class DiabloVolumeManager(VolumeManager):
    """
    The volumes extension deployed by HP Cloud uses an older endpoint,
    ``/os-volumes``, from the OpenStack Diablo timeframe.  Unfortunately, the
    endpoint in ``python-novaclient`` is hardcoded.  This class overrides any
    methods that specify an endpoint of ``/volumes``.

    Besides the use of the ``/os-volumes`` endpoint, the methods in this class
    were mostly pinched from its base,
    :class:`novaclient.v1_1.volumes.VolumeManager`.

    """
    def create(self, size, snapshot_id=None, display_name=None,
            display_description=None, volume_type=None, availability_zone=None,
            imageRef=None):
        body = {
                'volume': {
                    'size': size,
                    'snapshot_id': snapshot_id,
                    'display_name': display_name,
                    'display_description': display_description,
                    'volume_type': volume_type,
                    'availability_zone': availability_zone,
                    'imageRef': imageRef,
                    }
                }
        return self._create(VOLUMES_ENDPOINT, body, 'volume')

    def get(self, volume_id):
        return self._get("%s/%s" % (VOLUMES_ENDPOINT, volume_id), "volume")

    def list(self, detailed=True, search_opts=None):
        path = VOLUMES_ENDPOINT + '/detail' if detailed else VOLUMES_ENDPOINT

        if search_opts:
            qparams = dict((k, v) for (k, v) in search_opts.iteritems() if v)
            url = '%s?%s' % (path, urllib.urlencode(qparams))
        else:
            url = path

        return self._list(url, "volumes")

    def delete(self, volume):
        self._delete("%s/%s" % (VOLUMES_ENDPOINT, base.getid(volume)))



# Copyright 2012 - John Calixto
#
# This file is part of bang.
#
# bang is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# bang is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with bang.  If not, see <http://www.gnu.org/licenses/>.
from collections import Callable
from ..util import log
from .. import BangError


class Deployer(object):
    """Base class for all deployers"""
    def __init__(self, stack, config):
        self.stack = stack
        self.phases = []
        self.inventory_phases = []
        for k, v in config.iteritems():
            if '-' in k or ' ' in k:
                raise BangError(
                        'No hyphens or spaces in config key names please! %s'
                        % k
                        )
            self.__dict__[k] = v

    def deploy(self):
        for should_run, action in self.phases:
            if isinstance(should_run, Callable):
                if should_run():
                    action()
            elif should_run:
                action()

    def inventory(self):
        """
        Gathers ansible inventory data.

        Looks for existing servers that are members of the stack.

        Does not attempt to *create* any resources.
        """
        for action in self.inventory_phases:
            action()

    def run(self, action):
        """
        Runs through the phases defined by :attr:`action`.

        :param str action:  Either ``deploy`` or ``inventory``.

        """
        deployer = self.__class__.__name__
        log.info('Running %s...' % deployer)
        try:
            if action == 'deploy':
                self.deploy()
            elif action == 'inventory':
                self.inventory()
        except BangError as e:
            log.error(e.message)
            raise
        log.info('%s complete.' % deployer)



# Copyright 2012 - John Calixto
#
# This file is part of bang.
#
# bang is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# bang is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with bang.  If not, see <http://www.gnu.org/licenses/>.
import pymysql
from novaclient.client import Client as NovaClient

from ... import attributes as A, resources as R
from ...util import log, poll_with_timeout
from ..openstack import (OpenStack, Nova, RedDwarf,
        DEFAULT_STORAGE_SIZE_GB, DEFAULT_TIMEOUT_S, db_to_dict)
from .nova_ext import DiabloVolumeManager
from .reddwarf import HPDbaas


def fix_hp_addrs(server):
    """
    Works around hpcloud's peculiar "all ip addresses are returned as private
    even though one is public" bug.  This is also what the official hpfog gem
    does in the ``Fog::Compute::HP::Server#public_ip_address`` method.

    :param dict server:  Contains the server ID, a list of public IP addresses,
        and a list of private IP addresses.

    """
    fixed = {A.server.ID: server[A.server.ID]}
    both = server.get(A.server.PRIVATE_IPS)
    if both:
        fixed[A.server.PUBLIC_IPS] = [both[1]]
        fixed[A.server.PRIVATE_IPS] = [both[0]]
    return fixed


class HPNova(Nova):
    def find_servers(self, *args, **kwargs):
        """
        Wraps :meth:`bang.providers.openstack.Nova.find_servers` to apply
        hpcloud specialization, namely pulling IP addresses from the hpcloud's
        non-standard return values.

        """
        servers = super(HPNova, self).find_servers(*args, **kwargs)
        return map(fix_hp_addrs, servers)

    def create_server(self, *args, **kwargs):
        """
        Wraps :meth:`bang.providers.openstack.Nova.create_server` to apply
        hpcloud specialization, namely pulling IP addresses from the hpcloud's
        non-standard return values.

        """
        # hpcloud's management console stuffs all of its tags in a "tags" tag.
        # populate it with the stack and role values here only at server
        # creation time.  what users do with it after server creation is up to
        # them.
        tags = kwargs['tags']
        tags[A.tags.TAGS] = ','.join([
                tags.get(A.tags.STACK, ''),
                tags.get(A.tags.ROLE, ''),
                ])
        s = super(HPNova, self).create_server(*args, **kwargs)
        return fix_hp_addrs(s)


class HPRedDwarf(RedDwarf):
    def create_db(self, instance_name, instance_type,
            admin_username, admin_password, security_groups=None, db_name=None,
            storage_size_gb=DEFAULT_STORAGE_SIZE_GB,
            timeout_s=DEFAULT_TIMEOUT_S):
        """
        Creates a database instance.

        This method blocks until the db instance is active, or until
        :attr:`timeout_s` has elapsed.

        By default, hpcloud *assigns* an automatically-generated set of
        credentials for an admin user.  In addition to launching the db
        instance, this method uses the autogenerated credentials to login to
        the server and create the intended admin user based on the credentials
        supplied as method arguments.

        :param str instance_name:  A name to assign to the db instance.

        :param str instance_type:  The server instance type (e.g. ``medium``).

        :param str admin_username:  The admin username.

        :param str admin_password:  The admin password.

        :param security_groups:  *Not used in hpcloud*.

        :param str db_name:  The database name.  If this is not specified, the
            database will be named the same as the :attr:`instance_name`.

        :param int storage_size_gb:  The size of the storage volume in GB.

        :param float timeout_s:  The number of seconds to poll for an active
            database server before failing.  This value is also used when
            attempting to connect to the running mysql server.

        :rtype:  :class:`dict`

        """
        db = self._create_db(instance_name, instance_type,
                storage_size_gb)

        # hang on to these... hpcloud only provides a way to generate a new
        # set of username/password - there is no way to retrieve the originals.
        default_creds = db.credential
        log.debug('Credentials for %s: %s' % (instance_name, default_creds))

        instance = self._poll_instance_status(db, timeout_s)

        # we're taking advantage of a security bug in hpcloud's dbaas security
        # group rules.  the default *security* is to allow connections from
        # everywhere in the world.
        def connect():
            try:
                return pymysql.connect(
                        host=instance.hostname,
                        port=instance.port,
                        # db=self.database,
                        user=default_creds['username'],
                        passwd=default_creds['password'],
                        connect_timeout=timeout_s,
                        )
            except:
                log.warn("Could not connect to db, %s" % instance_name)
                # log.debug("Connection exception", exc_info=True)

        log.info("Connecting to %s..." % instance_name)
        db = poll_with_timeout(timeout_s, connect, 10)
        cur = db.cursor()
        cur.execute(
                "grant all privileges on *.* "
                "to '%s'@'%%' identified by '%s' "
                "with grant option"
                % (admin_username, admin_password)
                )
        cur.execute("flush privileges")
        return db_to_dict(instance)


class HPCloud(OpenStack):

    REDDWARF_SERVICE_TYPE = 'hpext:dbaas'
    REDDWARF_CLIENT_CLASS = HPDbaas

    def __init__(self, *args, **kwargs):
        super(HPCloud, self).__init__(*args, **kwargs)
        cm = self.CONSUL_MAP
        cm[R.SERVERS] = HPNova
        cm[R.DATABASES] = HPRedDwarf

    def _get_nova_client(self):
        args = self.get_nova_client_args()
        if 'access_key_id' in self.creds and 'secret_access_key' in self.creds:
            # this pluggable auth requires api version 2
            args[0] = '2'

        kwargs = self.get_nova_client_kwargs()
        if 'access_key_id' in self.creds and 'secret_access_key' in self.creds:
            kwargs['auth_system'] = 'secretkey'

        nc = NovaClient(*args, **kwargs)
        nc.volumes = DiabloVolumeManager(nc)
        return nc

    def authenticate(self):
        """
        Authenticate against the HP Cloud Identity Service.  This is the first
        step in any hpcloud.com session, although this method is automatically
        called when accessing higher-level methods/attributes.

        **Examples of Credentials Configuration**

        - Bare minimum for authentication using HP API keys::

            deployer_credentials:
              hpcloud:
                auth_url: https://region-a.geo-1.identity.hpcloudsvc.com:35357/v2.0/
                tenant_name: farley.mowat-tenant1
                access_key_id: MZOFIE9S83FOS248FIE3
                secret_access_key: EU859vjksor73gkY378f9gkslbkrabcxwfyW2loo

        - With multiple *compute* availability zones activated, the region must
          also be specified (due to current limitations in the OpenStack client
          libraries)::

            deployer_credentials:
              hpcloud:
                auth_url: https://region-a.geo-1.identity.hpcloudsvc.com:35357/v2.0/
                tenant_name: farley.mowat-tenant1
                access_key_id: MZOFIE9S83FOS248FIE3
                secret_access_key: EU859vjksor73gkY378f9gkslbkrabcxwfyW2loo
                region_name: az-1.region-a.geo-1

        - Using ``username`` and ``password`` is also allowed, but
          discouraged::

            deployer_credentials:
              hpcloud:
                auth_url: https://region-a.geo-1.identity.hpcloudsvc.com:35357/v2.0/
                tenant_name: farley.mowat-tenant1
                username: farley.mowat
                password: NeverCryW0lf

        When both API keys and ``username+password`` are specified, the API
        keys are used.

        """
        log.info("Authenticating to HP Cloud...")
        creds = self.creds
        access_key_id = creds.get('access_key_id', '')
        secret_access_key = creds.get('secret_access_key', '')
        # prefer api key + secret key, but fallback to username + password
        if access_key_id and secret_access_key:
            self.nova_client.client.os_access_key_id = access_key_id
            self.nova_client.client.os_secret_key = secret_access_key
        self.nova_client.authenticate()



# Copyright 2012 - John Calixto
#
# This file is part of bang.
#
# bang is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# bang is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with bang.  If not, see <http://www.gnu.org/licenses/>.
from .hpcloud import HPCloud
from .aws import AWS


PROVIDER_MAP = {
        'hpcloud': HPCloud,
        'aws': AWS,
        }

# provider object cache:
_PROVIDERS = {}


def get_provider(name, creds):
    """
    Generates and memoizes a :class:`~bang.providers.provider.Provider` object
    for the given name.

    :param str name:  The provider name, as given in the config stanza.  This
        token is used to find the
        appropriate :class:`~bang.providers.provider.Provider`.

    :param dict creds:  The credentials dictionary that is appropriate for the
        desired provider.  Typically, a sub-dict from the main stack config.

    :rtype:  :class:`~bang.providers.provider.Provider`

    """
    p = _PROVIDERS.get(name)
    if not p:
        provider = PROVIDER_MAP.get(name)
        p = provider(creds)
        _PROVIDERS[name] = p
    return p



# Copyright 2012 - John Calixto
#
# This file is part of bang.
#
# bang is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# bang is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with bang.  If not, see <http://www.gnu.org/licenses/>.
"""
Base classes and definitions for bang deployers (deployable components)
"""
from . import cloud
from .. import resources as R
from ..providers import get_provider
from ..util import log


DEPLOYER_MAP = {
        R.SSH_KEYS: cloud.SSHKeyDeployer,
        R.SERVERS: cloud.ServerDeployer,
        R.SERVER_SECURITY_GROUPS: cloud.SecurityGroupDeployer,
        R.SERVER_SECURITY_GROUP_RULES: cloud.SecurityGroupRulesetDeployer,
        R.BUCKETS: cloud.BucketDeployer,
        R.DATABASES: cloud.DatabaseDeployer,
        }


def get_stage_deployers(keys, stack):
    """
    Returns a list of deployer objects that *create* cloud resources.  Each
    member of the list is responsible for provisioning a single stack resource
    (e.g. a virtual server, a security group, a bucket, etc...).

    :param keys:  A list of top-level configuration keys for which to create
        deployers.
    :type keys:  :class:`~collections.Iterable`

    :param config:  A stack object.
    :type config:  :class:`~bang.stack.Stack`

    :rtype:  :class:`list` of :class:`~bang.deployers.deployer.Deployer`

    """
    config = stack.config
    creds = config['deployer_credentials']
    deployers = []
    for res_type in keys:
        res_configs = config.get(res_type)
        if not res_configs:
            continue
        log.debug("Found config for resource type, %s" % res_type)
        for res_config in res_configs:
            pname = res_config['provider']
            provider = get_provider(pname, creds[pname])
            consul = provider.get_consul(res_type)
            if not consul:
                log.warn("%s does not provide %s" % (pname, res_type))
                continue
            deployer = DEPLOYER_MAP[res_type]
            count = res_config.get('instance_count', 1)
            ds = [deployer(stack, res_config, consul) for _ in range(count)]
            deployers.extend(ds)
    return deployers



# Copyright 2012 - John Calixto
#
# This file is part of bang.
#
# bang is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# bang is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with bang.  If not, see <http://www.gnu.org/licenses/>.
import time
from .. import attributes as A
from ..util import log
from .deployer import Deployer


class BaseDeployer(Deployer):
    """Base class for all cloud resource deployers"""
    def __init__(self, stack, config, consul):
        super(BaseDeployer, self).__init__(stack, config)
        self._consul = consul

    @property
    def consul(self):
        return self._consul


class RegionedDeployer(BaseDeployer):
    """Deployer that automatically sets its region"""
    @property
    def consul(self):
        self._consul.set_region(self.region_name)
        return self._consul


class SSHKeyDeployer(RegionedDeployer):
    """
    Registers SSH keys with cloud providers so they can be used at
    server-launch time.

    """
    def __init__(self, *args, **kwargs):
        super(SSHKeyDeployer, self).__init__(*args, **kwargs)
        self.found = False
        self.phases = [
                (True, self.find_existing),
                (lambda: not self.found, self.register),
                ]

    def find_existing(self):
        """Searches for an existing SSH key matching the name."""
        self.found = self.consul.find_ssh_pub_key(self.name)

    def register(self):
        """Registers SSH key with provider."""
        log.info('Installing ssh key, %s' % self.name)
        self.consul.create_ssh_pub_key(self.name, self.key)


class ServerDeployer(RegionedDeployer):

    def __init__(self, *args, **kwargs):
        super(ServerDeployer, self).__init__(*args, **kwargs)
        self.namespace = self.stack.get_namespace(self.name)
        self.server_attrs = None
        self.phases = [
                (True, self.find_existing),
                (lambda: not self.server_attrs, self.create),
                (True, self.add_to_inventory),
                ]
        self.inventory_phases = [
                self.find_existing,
                self.add_to_inventory,
                ]

    def find_existing(self):
        """
        Searches for existing server instances with matching tags.  To match,
        the existing instances must also be "running".

        """
        instances = self.consul.find_servers(self.tags)
        maxnames = len(instances)
        while instances:
            i = instances.pop(0)
            server_id = i[A.server.ID]
            if self.namespace.add_if_unique(server_id):
                log.info('Found existing server, %s' % server_id)
                self.server_attrs = i
                break
            if len(self.namespace.names) >= maxnames:
                break
            instances.append(i)

    def create(self):
        """Launches a new server instance."""
        self.server_attrs = self.consul.create_server(
                "%s-%s" % (self.stack.name, self.name),
                self.disk_image_id,
                self.instance_type,
                self.ssh_key_name,
                tags=self.tags,
                availability_zone=self.availability_zone,
                timeout_s=self.launch_timeout_s,
                security_groups=self.security_groups,
                )
        log.debug('Post launch delay: %d s' % self.post_launch_delay_s)
        time.sleep(self.post_launch_delay_s)

    def add_to_inventory(self):
        """Adds host to stack inventory"""
        if not self.server_attrs:
            return
        for addy in self.server_attrs[A.server.PUBLIC_IPS]:
            self.stack.add_host(addy, self.groups, self.hostvars)


class SecurityGroupDeployer(RegionedDeployer):
    def __init__(self, *args, **kwargs):
        super(SecurityGroupDeployer, self).__init__(*args, **kwargs)
        self.group = None
        self.phases = [
                (True, self.find_existing),
                (lambda: not self.group, self.create),
                ]
        self.attrs = {}

    def find_existing(self):
        """Finds existing secgroup"""
        self.group = self.consul.find_secgroup(self.name)

    def create(self):
        """Creates a new security group"""
        self.consul.create_secgroup(self.name, self.description)


class SecurityGroupRulesetDeployer(RegionedDeployer):
    def __init__(self, *args, **kwargs):
        super(SecurityGroupRulesetDeployer, self).__init__(*args, **kwargs)
        self.create_these_rules = []
        self.delete_these_rules = []
        self.phases = [
                (True, self.find_existing),
                (lambda: self.create_these_rules or self.delete_these_rules,
                    self.apply_rule_changes),
                ]

    def find_existing(self):
        """
        Finds existing rule in secgroup.

        Populates ``self.create_these_rules`` and ``self.delete_these_rules``.

        """
        sg = self.consul.find_secgroup(self.name)
        current = {}
        for rule in sg.rules:
            src_group = rule.get('group')
            # TODO: allow for groups with different tenant ids.
            if src_group:
                src = src_group['name']
            else:
                src = rule['ip_range'].get('cidr', '')
            parsed = (
                    rule.get('ip_protocol'),
                    rule.get('from_port'),
                    rule.get('to_port'),
                    src,
                    )
            current[parsed] = rule['id']

        exp_rules = []
        for rule in self.rules:
            exp = (
                    rule[A.secgroup.PROTOCOL],
                    rule[A.secgroup.FROM],
                    rule[A.secgroup.TO],
                    rule[A.secgroup.SOURCE],
                    )
            exp_rules.append(exp)
            if exp in current:
                del current[exp]
            else:
                self.create_these_rules.append(exp)

        self.delete_these_rules.extend(current.itervalues())

        log.debug('Create these rules: %s' % self.create_these_rules)
        log.debug('Delete these rules: %s' % self.delete_these_rules)

    def apply_rule_changes(self):
        """
        Makes the security group rules match what is defined in the Bang
        config file.

        """
        # TODO: add error handling
        for rule in self.create_these_rules:
            args = rule + (self.name, )
            self.consul.create_secgroup_rule(*args)
            log.info("Authorized: %s" % str(rule))
        for rule_id in self.delete_these_rules:
            self.consul.delete_secgroup_rule(rule_id)
            log.info("Revoked: %s" % rule_id)


class BucketDeployer(BaseDeployer):
    def __init__(self, *args, **kwargs):
        super(BucketDeployer, self).__init__(*args, **kwargs)
        self.phases = [
                (True, self.create),
                ]

    def create(self):
        """Creates a new bucket"""
        self.consul.create_bucket("%s-%s" % (self.stack.name, self.name))


class DatabaseDeployer(BaseDeployer):
    def __init__(self, *args, **kwargs):
        super(DatabaseDeployer, self).__init__(*args, **kwargs)
        self.instance_name = "%s-%s" % (self.stack.name, self.name)
        self.db_attrs = None
        self.phases = [
                (True, self.find_existing),
                (lambda: not self.db_attrs, self.create),
                (True, self.add_to_inventory),
                ]
        self.inventory_phases = [
                self.find_existing,
                self.add_to_inventory,
                ]

    def find_existing(self):
        """
        Searches for existing db instance with matching name.  To match, the
        existing instance must also be "running".

        """
        self.db_attrs = self.consul.find_db_instance(self.instance_name)

    def create(self):
        """Creates a new database"""
        self.db_attrs = self.consul.create_db(
                self.instance_name,
                self.instance_type,
                self.admin_username,
                self.admin_password,
                db_name=self.db_name,
                storage_size_gb=self.storage_size,
                timeout_s=self.launch_timeout_s,
                )

    def add_to_inventory(self):
        """Adds db host to stack inventory"""
        host = self.db_attrs.pop(A.database.HOST)
        self.stack.add_host(
                host,
                self.groups,
                self.db_attrs
                )



# Copyright 2012 - John Calixto
#
# This file is part of bang.
#
# bang is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# bang is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with bang.  If not, see <http://www.gnu.org/licenses/>.
import boto
import boto.ec2
from boto.exception import EC2ResponseError
import time

from .. import TimeoutError, resources as R, attributes as A
from ..util import log, poll_with_timeout
from .bases import Provider, Consul


DEFAULT_TIMEOUT_S = 120


def server_to_dict(server):
    """
    Returns the :class:`dict` representation of a server object.

    The returned :class:`dict` is meant to be consumed by
    :class:`~bang.deployers.cloud.ServerDeployer` objects.

    """
    return {
            A.server.ID: server.id,
            A.server.PUBLIC_IPS: [server.public_dns_name],
            A.server.PRIVATE_IPS: [server.private_dns_name],
            }


class EC2(Consul):
    """The consul for the compute service in AWS (EC2)."""
    def __init__(self, *args, **kwargs):
        super(EC2, self).__init__(*args, **kwargs)
        creds = self.provider.creds
        self.access_key_id = creds[A.creds.ACCESS_KEY_ID]
        self.secret_key = creds[A.creds.SECRET_ACCESS_KEY]

        # this connection lets boto pick the default region.  be sure to use
        # set_region() if you need a specific region.
        self.ec2 = boto.connect_ec2(self.access_key_id, self.secret_key)

    def set_region(self, region_name):
        log.debug("Setting region to %s" % region_name)
        self.ec2 = boto.ec2.connect_to_region(
                region_name,
                aws_access_key_id=self.access_key_id,
                aws_secret_access_key=self.secret_key,
                )

    def find_servers(self, tags, running=True):
        """
        Returns any servers in the region that have tags that match the
        key-value pairs in :attr:`tags`.

        :param Mapping tags:  A mapping object in which the keys are the tag
            names and the values are the tag values.

        :param bool running:  A flag to limit server list to instances that are
            actually *running*.

        :rtype:  :class:`list` of :class:`dict` objects.  Each :class:`dict`
            describes a single server instance.

        """
        filters = dict([('tag:%s' % key, val) for key, val in tags.items()])
        if running:
            filters['instance-state-name'] = 'running'

        res = self.ec2.get_all_instances(filters=filters)
        instances = [server_to_dict(i) for r in res for i in r.instances]
        log.debug('instances: %s' % instances)
        return instances

    def create_server(self, basename, disk_image_id, instance_type,
            ssh_key_name, tags=None, availability_zone=None,
            timeout_s=DEFAULT_TIMEOUT_S, **kwargs):
        """
        Creates a new server instance.  This call blocks until the server is
        created and available for normal use, or :attr:`timeout_s` has elapsed.

        :param str basename:  An identifier for the server.  A random postfix
            will be appended to this basename to work around OpenStack Nova
            REST API limitations.

        :param str disk_image_id:  The identifier of the base disk image to use
            as the rootfs.

        :param str instance_type:  The name of an OpenStack instance type, or
            *flavor*.  This is specific to the OpenStack provider installation.

        :param str ssh_key_name:  The name of the ssh key to inject into the
            target server's ``authorized_keys`` file.  The key must already
            have been registered with the OpenStack Nova provider.

        :param tags:  Up to 5 key-value pairs of arbitrary strings to use as
            *tags* for the server instance.
        :type tags:  :class:`Mapping`

        :param str availability_zone:  The name of the availability zone in
            which to place the server.

        :param float timeout_s:  The number of seconds to poll for an active
            server before failing.  Defaults to ``0`` (i.e. Expect server to be
            active immediately).

        :rtype:  :class:`dict`

        """
        log.info(
                'Launching server %s... this could take a while...'
                % basename
                )
        res = self.ec2.run_instances(
                disk_image_id,
                instance_type=instance_type,
                key_name=ssh_key_name,
                placement=availability_zone,
                disable_api_termination=True,
                **kwargs
                )
        instance = res.instances[0]

        # we're too fast for EC2... slow down a little bit, twice
        time.sleep(2)

        def apply_tags():
            try:
                for key, val in tags.items():
                    instance.add_tag(key, val)
                return True
            except EC2ResponseError:
                pass
        if not poll_with_timeout(timeout_s, apply_tags, 5):
            raise TimeoutError('Could not tag server %s' % instance.id)

        def find_running_instance():
            if instance.update() == 'running':
                return instance
        running = poll_with_timeout(timeout_s, find_running_instance, 5)
        if not running:
            raise TimeoutError('Could not launch server within allotted time.')
        return server_to_dict(running)


class S3(Consul):
    pass


class RDS(Consul):
    pass


class AWS(Provider):

    CONSUL_MAP = {
            R.SERVERS: EC2,
            R.SERVER_SECURITY_GROUPS: EC2,
            R.SERVER_SECURITY_GROUP_RULES: EC2,
            R.BUCKETS: S3,
            R.DATABASES: RDS,
            }



# Copyright 2012 - John Calixto
#
# This file is part of bang.
#
# bang is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# bang is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with bang.  If not, see <http://www.gnu.org/licenses/>.
QUEUES = 'queues'
BUCKETS = 'buckets'
DATABASE_CREDS = 'database_credentials'
DATABASE_SECURITY_GROUPS = 'database_security_groups'
DATABASE_SECURITY_GROUP_RULES = 'database_security_group_rules'
DATABASES = 'databases'
SERVER_SECURITY_GROUPS = 'server_security_groups'
SERVER_SECURITY_GROUP_RULES = 'server_security_group_rules'
SERVER_COMMON_ATTRIBUTES = 'server_common_attributes'
SERVERS = 'servers'
SSH_KEYS = 'ssh_pub_keys'

# This is where the inter-resource dependencies are resolved.  Keep it simple
# until it needs to be more complicated.
#
# Each tuple in the list defines a *stage* - that is, a set of resources that
# can be deployed in parallel.
#
# The stack deployer starts deploying resources in the first tuple, waits for
# all of the resources to be deployed successfully, then moves on to the next
# tuple of resources, etc...  It always waits for all of the deployers in a
# stage/tuple to complete before moving to the next stage/tuple.
#
# If any resource deployment within a stage/tuple is *not* successful, the
# stack deployer does *not* proceed to the next stage - the deployment is
# terminated, and the errors are reported.
STAGES = [
        (
            DATABASE_SECURITY_GROUPS,
            SERVER_SECURITY_GROUPS,
            QUEUES,
            BUCKETS,
            SSH_KEYS,
            ),

        (
            SERVER_SECURITY_GROUP_RULES,
            ),

        (
            DATABASES,
            SERVERS,
            ),

        (
            DATABASE_SECURITY_GROUP_RULES,
            ),
        ]

CONVENIENCE_KEYS = [
        SERVER_COMMON_ATTRIBUTES,
        DATABASE_CREDS,
        ]

DYNAMIC_RESOURCE_KEYS = [k for s in STAGES for k in s] + CONVENIENCE_KEYS



# Copyright 2012 - John Calixto
#
# This file is part of bang.
#
# bang is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# bang is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with bang.  If not, see <http://www.gnu.org/licenses/>.
PUBLIC_IPS = 'public_ip_addresses'
PRIVATE_IPS = 'private_ip_addresses'
SECGROUPS = 'security_groups'
STACK_SECGROUPS = 'stack_security_groups'
EXTRA_SECGROUPS = 'extra_security_groups'
TAGS = 'tags'
NAME = 'name'
ID = 'id'
SCOPES = 'config_scopes'
VARS = 'hostvars'
GROUPS = 'groups'
REGION = 'region_name'
AZ = 'availability_zone'
LAUNCH_TIMEOUT = 'launch_timeout_s'
POST_DELAY = 'post_launch_delay_s'
SSH_KEY = 'ssh_key_name'
PROVIDER = 'provider'

# these are ansible magic vars
INV_NAME = 'inventory_hostname'
INV_NAME_SHORT = 'inventory_hostname_short'
GROUP_NAMES = 'group_names'



#!/usr/bin/env python
from setuptools import setup, find_packages
from bang import VERSION
import os.path


ETC = os.path.join(os.path.dirname(__file__), 'etc')

with open(os.path.join(ETC, 'requirements.pip')) as f:
    reqs = [l.strip() for l in f if '://' not in l]
reqs.append('distribute')

setup(
        name='bang',
        version=VERSION,
        author='fr33jc',
        author_email='fr33jc@gmail.com',
        packages=find_packages(exclude=['tests']),
        license='GPLv3',
        description='Server and cloud resource deployment automation',
        platforms='POSIX',
        url='https://github.com/fr33jc/bang',
        install_requires=reqs,
        scripts=['bin/bang'],
        )



# Copyright 2012 - John Calixto
#
# This file is part of bang.
#
# bang is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# bang is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with bang.  If not, see <http://www.gnu.org/licenses/>.
"""
Constants for attribute names of the various resources.

Just a place to stash more magic strings.
"""
from . import creds, server, secgroup, tags, database, ssh_key

# stack attributes
NAME = 'name'
VERSION = 'version'
LOGGING = 'logging'
PLAYBOOKS = 'playbooks'
STACK = 'stack'
DEPLOYER_CREDS = 'deployer_credentials'
KEY = 'key'



# Copyright 2012 - John Calixto
#
# This file is part of bang.
#
# bang is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# bang is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with bang.  If not, see <http://www.gnu.org/licenses/>.
from functools import wraps
from novaclient.client import Client as NovaClient
from swiftclient.client import Connection as SwiftConn
from reddwarfclient import Dbaas

from ... import BangError, TimeoutError, resources as R, attributes as A
from ...util import log, poll_with_timeout
from ..bases import Provider, Consul


DEFAULT_TIMEOUT_S = 120
DEFAULT_STORAGE_SIZE_GB = 20


def server_to_dict(server):
    """
    Returns the :class:`dict` representation of a server object.

    The returned :class:`dict` is meant to be consumed by
    :class:`~bang.deployers.cloud.ServerDeployer` objects.

    """
    addresses = server.addresses
    pub = addresses.get('public', [])
    priv = addresses.get('private', [])
    return {
            A.server.ID: server.id,
            A.server.PUBLIC_IPS: [a['addr'] for a in pub],
            A.server.PRIVATE_IPS: [a['addr'] for a in priv],
            }


def db_to_dict(db):
    """
    Returns the :class:`dict` representation of a database object.

    The returned :class:`dict` is meant to be consumed by
    :class:`~bang.deployers.cloud.ServerDeployer` objects.

    """
    return {
            A.database.HOST: db.hostname,
            A.database.PORT: db.port,
            }


class Nova(Consul):
    """The consul for the OpenStack compute service."""
    def __init__(self, *args, **kwargs):
        super(Nova, self).__init__(*args, **kwargs)
        self.nova = self.provider.nova_client

    def set_region(self, region_name):
        client = self.nova.client
        management_url = client.service_catalog.url_for(
                attr='region',
                filter_value=region_name,
                service_type=client.service_type,
                )
        client.set_management_url(management_url.rstrip('/'))

    def find_ssh_pub_key(self, name):
        """
        Returns ``True`` if an SSH key named :attr:`name` is found.

        Otherwise returns ``False``.

        :rtype:  :class:`bool`

        """
        return bool(self.nova.keypairs.findall(name=name))

    def create_ssh_pub_key(self, name, key):
        """
        Installs the public SSH key under the name :attr:`name`.

        Once installed, the key can be referenced when creating new server
        instances.

        """
        self.nova.keypairs.create(name, key)

    def find_servers(self, tags, running=True):
        """
        Returns any servers in the region that have tags that match the
        key-value pairs in :attr:`tags`.

        :param Mapping tags:  A mapping object in which the keys are the tag
            names and the values are the tag values.

        :param bool running:  A flag to limit server list to instances that are
            actually *running*.

        :rtype:  :class:`list` of :class:`dict` objects.  Each :class:`dict`
            describes a single server instance.

        """
        log.debug("Finding instances with tags: %s" % tags)
        search_opts = {}
        if running:
            search_opts['status'] = 'ACTIVE'
        all_servers = self.nova.servers.list(search_opts=search_opts)
        servers = []
        for s in all_servers:
            md = s.metadata
            mismatches = [k for k, v in tags.items() if v != md.get(k)]
            if mismatches:
                continue
            servers.append(server_to_dict(s))
        return servers

    def create_server(self, basename, disk_image_id, instance_type,
            ssh_key_name, tags=None, availability_zone=None,
            timeout_s=DEFAULT_TIMEOUT_S, **kwargs):
        """
        Creates a new server instance.  This call blocks until the server is
        created and available for normal use, or :attr:`timeout_s` has elapsed.

        :param str basename:  An identifier for the server.  A random postfix
            will be appended to this basename to work around OpenStack Nova
            REST API limitations.

        :param str disk_image_id:  The identifier of the base disk image to use
            as the rootfs.

        :param str instance_type:  The name of an OpenStack instance type, or
            *flavor*.  This is specific to the OpenStack provider installation.

        :param str ssh_key_name:  The name of the ssh key to inject into the
            target server's ``authorized_keys`` file.  The key must already
            have been registered with the OpenStack Nova provider.

        :param tags:  Up to 5 key-value pairs of arbitrary strings to use as
            *tags* for the server instance.
        :type tags:  :class:`Mapping`

        :param str availability_zone:  The name of the availability zone in
            which to place the server.

        :param float timeout_s:  The number of seconds to poll for an active
            server before failing.  Defaults to ``0`` (i.e. Expect server to be
            active immediately).

        :rtype:  :class:`dict`

        """
        nova = self.nova
        name = self.provider.gen_component_name(basename)
        log.info('Launching server %s... this could take a while...' % name)
        flavor = nova.flavors.find(name=instance_type)
        server = nova.servers.create(
                name,
                disk_image_id,
                flavor,
                key_name=ssh_key_name,
                meta=tags,
                availability_zone=availability_zone,
                **kwargs
                )

        def find_active():
            s = nova.servers.get(server.id)
            if s and s.status == 'ACTIVE':
                return s

        instance = poll_with_timeout(timeout_s, find_active, 5)
        if not instance:
            raise TimeoutError(
                    'Server %s failed to launch within allotted time.'
                    % server.id
                    )
        return server_to_dict(instance)

    def find_secgroup(self, name):
        """
        Searches for, and returns a
        :class:`novaclient.v1_1.security_groups.SecurityGroup` named
        :attr:`name`.
        """
        groups = self.nova.security_groups.findall(name=name)
        if not groups:
            return
        # TODO: will this ever happen?
        if len(groups) > 1:
            log.warn("Found too many groups named %s" % name)
        log.debug("... found group %s" % name)
        return groups.pop(0)

    def create_secgroup(self, name, desc):
        """
        Creates a new server security group.

        :param str name:  The name of the security group to create.
        :param str desc:  A short description of the group.

        """
        self.nova.security_groups.create(name, desc)

    def create_secgroup_rule(self, protocol, from_port, to_port,
            source, target):
        """
        Creates a new server security group rule.

        :param str protocol:  E.g. ``tcp``, ``icmp``, etc...
        :param int from_port:  E.g. ``1``
        :param int to_port:  E.g. ``65535``
        :param str source:
        :param str target:

        """
        nova = self.nova

        def get_id(gname):
            sg = nova.security_groups.find(name=gname)
            if not sg:
                raise BangError("Security group not found, %s" % gname)
            return str(sg.id)

        kwargs = {
                'ip_protocol': protocol,
                'from_port': str(from_port),
                'to_port': str(to_port),
                'parent_group_id': get_id(target),
                }
        if '/' in source:
            kwargs['cidr'] = source
        else:
            kwargs['group_id'] = get_id(source)
            # not sure if this is an openstack hack or an hpcloud hack, but
            # this is definitely required to get it working on hpcloud:
            kwargs['cidr'] = 'null'
        nova.security_group_rules.create(**kwargs)

    def delete_secgroup_rule(self, rule_id):
        """Deletes the security group rule identified by :attr:`rule_id`"""
        self.nova.security_group_rules.delete(rule_id)


class Swift(Consul):
    def find_buckets(self, prefix):
        _, buckets = self.provider.swift_client.get_account(prefix=prefix)
        # TODO: standardize the return value across providers
        return buckets

    def create_bucket(self, name, headers=None):
        """
        Creates a bucket named :attr:`name`.

        :param dict headers:  Any headers to use when performing the HTTP PUT.

        """
        self.provider.swift_client.put_container(name, headers)


class RedDwarf(Consul):
    def find_db_instance(self, name, running=True):
        """
        Searches for a db instance named :attr:`name`.

        :param str name:  The name of the target db instance.

        :param bool running:  A flag to only look for instances that are
            actually *running*.

        :rtype:  :class:`dict`

        """
        found = None
        for i in self.provider.reddwarf_client.instances.list():
            if i.name == name:
                if not running:  # don't care if it's running or not
                    found = i
                if i.status == 'running':
                    found = i
        if found:
            log.info("Found existing db, %s (%s)" % (found.name, found.id))
            return db_to_dict(found)

    def _create_db(self, instance_name, instance_type,
            storage_size_gb):
        rd = self.provider.reddwarf_client
        flavor = rd.flavors.find(name=instance_type)
        log.info('Launching db server %s...' % instance_name)
        # TODO:  Upstream RedDwarf has some notion of ``databases`` and
        # ``users``, both of which are optional args to the create() call
        # below.  Figure out what that means in practice.
        return rd.instances.create(
                instance_name,
                flavor.links[0]['href'],
                {'size': storage_size_gb}
                )

    def _poll_instance_status(self, db, timeout_s):
        log.info('Polling for db status...')

        def find_active():
            i = self.provider.reddwarf_client.instances.get(db.id)
            if i and i.status == 'running':
                return i

        instance = poll_with_timeout(timeout_s, find_active, 20)
        if not instance:
            raise TimeoutError(
                    'DB %s failed to launch within allotted time.' % db.id
                    )
        return instance

    def create_db(self, instance_name, instance_type, admin_username,
            admin_password, security_groups=None, db_name=None,
            storage_size_gb=DEFAULT_STORAGE_SIZE_GB,
            timeout_s=DEFAULT_TIMEOUT_S):
        """
        Creates a database instance.

        This method blocks until the db instance is active, or until
        :attr:`timeout_s` has elapsed.

        :param str instance_name:  A name to assign to the db instance.

        :param str instance_type:  The server instance type (e.g. ``medium``).

        :param str admin_username:  The admin username.

        :param str admin_password:  The admin password.

        :param security_groups:  A list of security groups in which to place
            the db instance.
        :type security_groups:  :class:`~collections.Iterable`

        :param str db_name:  The database name.  If this is not specified, the
            database will be named the same as the :attr:`instance_name`.

        :param int storage_size_gb:  The size of the storage volume in GB.

        :param float timeout_s:  The number of seconds to poll for an active
            database server before failing.

        :rtype:  :class:`dict`

        """
        # TODO: investigate what upstream RedDwarf does for admin users
        return db_to_dict(
                self._poll_instance_status(
                    self._create_db(instance_name, instance_type,
                        storage_size_gb),
                    timeout_s
                    )
                )


def authenticated(f):
    """Decorator that authenticates to Keystone automatically."""
    @wraps(f)
    def new_f(self, *args, **kwargs):
        if not self.nova_client.client.auth_token:
            self.authenticate()
        return f(self, *args, **kwargs)
    return new_f


class OpenStack(Provider):

    CONSUL_MAP = {
            R.SERVERS: Nova,
            R.SSH_KEYS: Nova,
            R.SERVER_SECURITY_GROUPS: Nova,
            R.SERVER_SECURITY_GROUP_RULES: Nova,
            R.BUCKETS: Swift,
            R.DATABASES: RedDwarf,
            }

    REDDWARF_SERVICE_TYPE = 'reddwarf'
    REDDWARF_CLIENT_CLASS = Dbaas

    def __init__(self, creds):
        super(OpenStack, self).__init__(creds)
        self._client = None
        self._swift = None
        self._reddwarf = None
        self.authenticate()

    @property
    @authenticated
    def os_auth_token(self):
        """Authentication token returned from Keystone."""
        return self.nova_client.client.auth_token

    @property
    @authenticated
    def os_catalog(self):
        """
        Service catalog returned from Keystone (Identity Service).

        This is specific to OpenStack providers.
        """
        return self.nova_client.client.service_catalog.catalog

    def _get_nova_client(self):
        args = self.get_nova_client_args()
        kwargs = self.get_nova_client_kwargs()
        return NovaClient(*args, **kwargs)

    @property
    def nova_client(self):
        """
        Each of the OpenStack client libraries (i.e. the ``python-*client``
        projects that live under https://github.com/openstack/) has its own way
        of connecting to the identity service (keystone).

        The object returned here is a :class:`novaclient.v1_1.client.Client`.

        """
        if not self._client:
            self._client = self._get_nova_client()
        return self._client

    @property
    def swift_client(self):
        if not self._swift:
            url = self.nova_client.client.service_catalog.url_for(
                    service_type='object-store'
                    )
            sconn = SwiftConn(
                    'usetoken',
                    'usetoken',
                    'usetoken',
                    preauthurl=url,
                    auth_version='2.0',
                    preauthtoken=self.os_auth_token
                    )
            self._swift = sconn
        return self._swift

    @property
    def reddwarf_client(self):
        if not self._reddwarf:
            url = self.nova_client.client.service_catalog.url_for(
                    service_type=self.REDDWARF_SERVICE_TYPE
                    )
            cli = self.REDDWARF_CLIENT_CLASS('usenova', 'usenova',
                    self.creds.get(A.creds.TENANT))
            cli.client.authenticate_with_token(self.os_auth_token, url)
            self._reddwarf = cli
        return self._reddwarf

    def get_nova_client_args(self):
        creds = self.creds
        # novaclient supports authentication plugins, so username + password
        # might not be the only set of credentials used to authenticate, so we
        # won't validate their existence here.  we do need tenant_name though.
        tenant_name = self.creds.get('tenant_name')
        if tenant_name:
            return [
                    '1.1',
                    creds.get('username', ''),
                    creds.get('password', ''),
                    tenant_name,
                    ]
        raise BangError("Need tenant_name to authenticate!")

    def get_nova_client_kwargs(self):
        creds = self.creds
        kwargs = dict((k, creds.get(k)) for k in ('auth_url', 'region_name'))
        # region_name is optional, but auth_url is required:
        if not kwargs['auth_url']:
            raise BangError("Need auth_url to authenticate!")
        return kwargs

    def authenticate(self):
        log.info("Authenticating to OpenStack...")
        self.nova_client.authenticate()
